mzirtfit.pml <- function(data, silent = TRUE, suppressWarnings = TRUE, 
                         ...) {
  data <- as.data.frame(data)
  stopifnot(c("x1", "x2", "x3", "x4", "x5", "age") %in% 
              names(data))
  age <- data$age
  
  ## require package
  stopifnot(require("lavaan"))
  
  data <- data[, c("x1", "x2", "x3", "x4", "x5")]
  
  ## fit model
  f05 <- paste(paste("lam", 1:5, "*", names(data)[1:5], 
                     sep = ""), collapse = " + ")
  f05 <- paste("theta =~ ", f05, sep = "")
  f06 <- paste(names(data)[1:5], " | th", 1:5, "*t1", 
               sep = "")
  f06 <- paste(f06, collapse = " \n ")
  f07 <- paste("alpha", 1:5, " := (lam", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f07 <- paste(f07, collapse = " \n ")
  f08 <- paste("beta", 1:5, " := (-th", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f08 <- paste(f08, collapse = " \n ")
  
  mod <- paste(f05, f06, f07, f08, sep = " \n ")
  writeLines(mod, "mod.txt")
  
  rval <- try(cfa(mod, data = data, std.lv = TRUE, 
                  ordered = names(data)[1:5], estimator = "PML"))
  return(rval)
}

mzgroupirtfit.pml <- function(data, silent = TRUE, 
                              suppressWarnings = TRUE, ...) {
  data <- as.data.frame(data)
  stopifnot(c("x1", "x2", "x3", "x4", "x5", "age") %in% 
              names(data))
  age <- data$age
  
  ## require package
  stopifnot(require("lavaan"))
  
  data <- data[, c("x1", "x2", "x3", "x4", "x5", 
                   "age")]
  
  ## fit model
  f05 <- paste(paste("lam", 1:5, "*", names(data)[1:5], 
                     sep = ""), collapse = " + ")
  f05 <- paste("theta =~ ", f05, sep = "")
  f06 <- paste(names(data)[1:5], " | th", 1:5, "*t1", 
               sep = "")
  f06 <- paste(f06, collapse = " \n ")
  f07 <- paste("alpha", 1:5, " := (lam", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f07 <- paste(f07, collapse = " \n ")
  f08 <- paste("beta", 1:5, " := (-th", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f08 <- paste(f08, collapse = " \n ")
  f09 <- paste(names(data)[1:5], " ~*~ 1*", names(data)[1:5], 
               sep = "", collapse = " \n ")
  f10 <- "theta ~~ c(1, NA, NA, NA, NA, NA, NA, NA)*theta"
  f11 <- "theta ~ c(0, NA, NA, NA, NA, NA, NA, NA)*1"
  simmodgroup <- paste(f05, f06, f07, f08, f09, f10, 
                       f11, sep = " \n ")
  writeLines(simmodgroup, "simmodgroup.txt")
  
  rval <- try(cfa(simmodgroup, data = data, std.lv = TRUE, 
                  ordered = names(data)[1:5], group = "age", 
                  group.equal = c("loadings", "thresholds"), 
                  estimator = "PML"))
  return(rval)
}

mzgroupmeanirtfit.pml <- function(data, silent = TRUE, 
                                  suppressWarnings = TRUE, ...) {
  data <- as.data.frame(data)
  stopifnot(c("x1", "x2", "x3", "x4", "x5", "age") %in% 
              names(data))
  age <- data$age
  
  ## require package
  stopifnot(require("lavaan"))
  
  data <- data[, c("x1", "x2", "x3", "x4", "x5", 
                   "age")]
  
  ## fit model
  f05 <- paste(paste("lam", 1:5, "*", names(data)[1:5], 
                     sep = ""), collapse = " + ")
  f05 <- paste("theta =~ ", f05, sep = "")
  f06 <- paste(names(data)[1:5], " | th", 1:5, "*t1", 
               sep = "")
  f06 <- paste(f06, collapse = " \n ")
  f07 <- paste("alpha", 1:5, " := (lam", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f07 <- paste(f07, collapse = " \n ")
  f08 <- paste("beta", 1:5, " := (-th", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f08 <- paste(f08, collapse = " \n ")
  f09 <- paste(names(data)[1:5], " ~*~ 1*", names(data)[1:5], 
               sep = "", collapse = " \n ")
  f10 <- "theta ~ c(0, NA, NA, NA, NA, NA, NA, NA)*1"
  f11 <- "theta ~~ c(1, 1, 1, 1, 1, 1, 1, 1)*theta"
  simmodgroupmean <- paste(f05, f06, f07, f08, f09, 
                           f10, f11, sep = " \n ")
  writeLines(simmodgroupmean, "simmodgroupmean.txt")
  
  rval <- try(cfa(simmodgroupmean, data = data, std.lv = TRUE, 
                  ordered = names(data)[1:5], group = "age", 
                  group.equal = c("loadings", "thresholds"), 
                  estimator = "PML"))
  return(rval)
}

mzgrconsirtfit.pml <- function(data, silent = TRUE, 
                               suppressWarnings = TRUE, ...) {
  data <- as.data.frame(data)
  stopifnot(c("x1", "x2", "x3", "x4", "x5", "age") %in% 
              names(data))
  age <- data$age
  
  ## require package
  stopifnot(require("lavaan"))
  
  data <- data[, c("x1", "x2", "x3", "x4", "x5", 
                   "age")]
  
  ## fit model
  f05 <- paste(paste("lam", 1:5, "*", names(data)[1:5], 
                     sep = ""), collapse = " + ")
  f05 <- paste("theta =~ ", f05, sep = "")
  f06 <- paste(names(data)[1:5], " | th", 1:5, "*t1", 
               sep = "")
  f06 <- paste(f06, collapse = " \n ")
  f07 <- paste("alpha", 1:5, " := (lam", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f07 <- paste(f07, collapse = " \n ")
  f08 <- paste("beta", 1:5, " := (-th", 1:5, ")/sqrt(1-lam", 
               1:5, "^2)", sep = "")
  f08 <- paste(f08, collapse = " \n ")
  f09 <- paste(names(data)[1:5], " ~*~ 1*", names(data)[1:5], 
               sep = "", collapse = " \n ")
  f10 <- "theta ~~ c(1, NA, NA, NA, NA, NA, NA, NA)*theta +
  c(NA, 'v', 'v', 'v', 'v', 'v', 'v', 'v')*theta "
  f11 <- "theta ~ c(0, NA, NA, NA, NA, NA, NA, NA)*1 +
  c(NA, 'm', 'm', 'm', 'm', 'm', 'm', 'm')*1"
  simmodgrcons <- paste(f05, f06, f07, f08, f09, 
                        f11, f10, sep = " \n ")
  writeLines(simmodgrcons, "simmodgrcons.txt")
  
  rval <- try(cfa(simmodgrcons, data = data, std.lv = TRUE, 
                  ordered = names(data)[1:5], group = "age", 
                  group.equal = c("loadings", "thresholds"), 
                  estimator = "PML"))
  return(rval)
}

estfun.lavaan <- pmlScores <- function(object) {
  library("lavaan")
  ## pull arguments value from objects shortcuts
  lavdata <- object@Data
  lavmodel <- object@Model
  lavsamplestats <- object@SampleStats
  lavfit <- object@Fit
  lavcache <- object@Cache
  
  ## number variables/sample size
  ntab <- unlist(lavsamplestats@nobs)
  ntot <- lavsamplestats@ntotal
  
  Score.mat <- matrix(NA, ntot, length(coef(object)))
  
  Sigma.hat <- lavaan:::computeSigmaHat(lavmodel = lavmodel)
  TH <- lavaan:::computeTH(lavmodel = lavmodel)
  if (lavmodel@nexo > 0L) {
    PI <- lavaan:::computePI(lavmodel = lavmodel)
  } else {
    PI <- vector("list", length = lavsamplestats@ngroups)
  }
  lpt <- parTable(object)
  eq.rows <- which(lpt$op == "==")
  eq.pars <- tapply(lpt$rhs[eq.rows], lpt$lhs[eq.rows], 
                    unique)
  fsname <- lpt$plabel[lpt$free > 0]
  
  for (g in 1:object@SampleStats@ngroups) {
    score.H1 <- lavaan:::pml_deriv1(Sigma.hat = Sigma.hat[[g]], 
                                    TH = TH[[g]], th.idx = lavmodel@th.idx[[g]], 
                                    num.idx = lavmodel@num.idx[[g]], 
                                    X = lavdata@X[[g]], 
                                    eXo = lavdata@eXo[[g]], PI = PI[[g]], 
                                    lavcache = lavcache[[g]], 
                                    missing = lavdata@missing, scores = TRUE, 
                                    negative = TRUE)
    Delta <- lavaan:::computeDelta(lavmodel)[[g]]
    wi <- lavdata@case.idx[[g]]
    Score.mat[wi, ] <- -score.H1 %*% Delta
    colnames(Score.mat) <- names(coef(object))
  }
  rmcols <- which(fsname %in% unlist(eq.pars))
  for (i in 1:length(eq.pars)) {
    sumcols <- which(fsname %in% eq.pars[[i]])
    eqloc <- match(names(eq.pars)[i], fsname)
    Score.mat[, eqloc] <- rowSums(Score.mat[, c(eqloc, 
                                                sumcols)])
  }
  Score.mat <- Score.mat[, -rmcols]
  Score.mat
}

ctr_pml_plrt_nested <- function(fit_objH0, fit_objH1) {
  
  # sanity check, perhaps we misordered H0 and H1 in
  # the function call??
  if (fit_objH1@test[[1]]$df > fit_objH0@test[[1]]$df) {
    tmp <- fit_objH0
    fit_objH0 <- fit_objH1
    fit_objH1 <- tmp
  }
  
  # check if we have equality constraints
  if (fit_objH0@Model@eq.constraints) {
    equalConstr = TRUE
  } else {
    equalConstr = FALSE
  }
  
  nsize <- fit_objH0@SampleStats@ntotal
  PLRT <- 2 * (fit_objH1@optim$logl - fit_objH0@optim$logl)
  
  # create a new object 'objH1_h0': the object 'H1',
  # but where the parameter values are from H0
  objH1_h0 <- lavaan:::lav_test_diff_m10(m1 = fit_objH1, 
                                         m0 = fit_objH0, test = FALSE)
  
  # EqMat
  EqMat <- lavaan:::lav_test_diff_A(m1 = fit_objH1, 
                                    m0 = fit_objH0)
  
  # Observed information (= for PML, this is Hessian
  # / N)
  Hes.theta0 <- lavTech(objH1_h0, "information.observed")
  
  # handle possible constraints in H1 (and therefore
  # also in objH1_h0)
  Inv.Hes.theta0 <- lavaan:::lav_model_information_augment_invert(lavmodel = objH1_h0@Model, 
                                                                  information = Hes.theta0, 
                                                                  inverted = TRUE)
  
  # the estimated variability matrix is given (=unit
  # information first order)
  J.theta0 <- lavaan:::lavTech(objH1_h0, "first.order")
  
  # the Inverse of the G matrix
  Inv.G <- Inv.Hes.theta0 %*% J.theta0 %*% Inv.Hes.theta0
  
  MInvGtM <- EqMat %*% Inv.G %*% t(EqMat)
  MinvHtM <- EqMat %*% Inv.Hes.theta0 %*% t(EqMat)
  ## ajust tol for computational stable.
  Inv_MinvHtM <- solve(MinvHtM, tol = 1e-17)
  tmp.prod <- MInvGtM %*% Inv_MinvHtM
  tmp.prod2 <- tmp.prod %*% tmp.prod
  sum.eig <- sum(diag(tmp.prod))
  sum.eigsq <- sum(diag(tmp.prod2))
  
  FSMA.PLRT <- (sum.eig/sum.eigsq) * PLRT
  adj.df <- (sum.eig * sum.eig)/sum.eigsq
  pvalue <- 1 - pchisq(FSMA.PLRT, df = adj.df)
  
  list(FSMA.PLRT = FSMA.PLRT, adj.df = adj.df, pvalue = pvalue)
}